﻿//-----------------------------------------------------------------------
// <copyright file="MainForm.cs" company="Tigerware Development">
//     Copyright(c) Tigerware Development. All rights reserved.
// </copyright>
//-----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Windows.Forms;

namespace ElexolIO24PTU.Net
{
    public partial class MainForm : Form
    {
        // delegates
        protected delegate void UpdateValues(byte value);

        // declarations
        bool _scanned;
        bool _ReadingEE;
        bool _EEWriteFlag;
        bool _EEWriteSuccess;
        bool _EEReadSuccess;
        bool _AutoLoadTarget;
        bool _SetPowerUp;
        int _EEReadDelay;

        byte[] _EEVal = new byte[128];

        byte _EEAddress;
        int _EEData;

        IPAddress _deviceIP;
        int _devicePort;
        IPEndPoint _deviceEP;

        UdpClient _udp;

        DebugConsole debug = new DebugConsole();

        public MainForm()
        {
            InitializeComponent();

            _EEReadDelay = 10;
            _scanned = false;
            _devicePort = 2424;
            _udp = new UdpClient(AddressFamily.InterNetwork);   // uses random outgoing port
        }

        #region PRIVATE

        /// <summary>
        /// load controls with EEPROM data (from module query or load config)
        /// </summary>
        private void loadControls()
        {
            // IP Address information
            tbFixedIP1.Text = _EEVal[12].ToString();
            tbFixedIP2.Text = _EEVal[13].ToString();
            tbFixedIP3.Text = _EEVal[14].ToString();
            tbFixedIP4.Text = _EEVal[15].ToString();

            tbMaskIP1.Text = _EEVal[50].ToString();
            tbMaskIP2.Text = _EEVal[51].ToString();
            tbMaskIP3.Text = _EEVal[52].ToString();
            tbMaskIP4.Text = _EEVal[53].ToString();

            tbGwayIP1.Text = _EEVal[54].ToString();
            tbGwayIP2.Text = _EEVal[55].ToString();
            tbGwayIP3.Text = _EEVal[56].ToString();
            tbGwayIP4.Text = _EEVal[57].ToString();

            tbProgrammablePort.Text = (_EEVal[58] + _EEVal[59] * 256).ToString();
            cbEnableFixedIP.Checked = (_EEVal[10] & 1) == 0;

            // Programming options
            // Direction
            cbProgramDirectionEnable.Checked = (_EEVal[10] & 2) == 0;
            updateProgramDirectionA(_EEVal[16]);
            updateProgramDirectionB(_EEVal[21]);
            updateProgramDirectionC(_EEVal[26]);

            // Values
            updateProgramValueA(_EEVal[17]);
            updateProgramValueB(_EEVal[22]);
            updateProgramValueC(_EEVal[27]);

            // Threshold
            updateProgramThresholdA(_EEVal[18]);
            updateProgramThresholdB(_EEVal[23]);
            updateProgramThresholdC(_EEVal[28]);

            // Pullup
            updateProgramPullupA(_EEVal[19]);
            updateProgramPullupB(_EEVal[24]);
            updateProgramPullupC(_EEVal[29]);

            // Schmitt
            updateProgramSchmittA(_EEVal[20]);
            updateProgramSchmittB(_EEVal[25]);
            updateProgramSchmittC(_EEVal[30]);

            // Autoscan
            cbAutoscanMaskEnable.Checked = (_EEVal[10] & 4) == 0;
            updateAutoscanMaskA(_EEVal[32]);
            updateAutoscanMaskB(_EEVal[33]);
            updateAutoscanMaskC(_EEVal[34]);

            // autoscan period and filter
            tbAutoscanFilterCount.Text = _EEVal[35].ToString();
            tbAutoscanScanPeriod.Text = (_EEVal[37] * 256 + _EEVal[36]).ToString();

            // autoscan target
            tbAutoscanIP1.Text = _EEVal[44].ToString();
            tbAutoscanIP2.Text = _EEVal[45].ToString();
            tbAutoscanIP3.Text = _EEVal[46].ToString();
            tbAutoscanIP4.Text = _EEVal[47].ToString();
            tbAutoscanTargetPort.Text = (_EEVal[49] * 256 + _EEVal[48]).ToString();

            tbAutoscanMAC1.Text = _EEVal[38].ToString("X2");
            tbAutoscanMAC2.Text = _EEVal[39].ToString("X2");
            tbAutoscanMAC3.Text = _EEVal[40].ToString("X2");
            tbAutoscanMAC4.Text = _EEVal[41].ToString("X2");
            tbAutoscanMAC5.Text = _EEVal[42].ToString("X2");
            tbAutoscanMAC6.Text = _EEVal[43].ToString("X2");

            // display EEPROM
            EEShow();

        }

        /// <summary>
        /// display eeprom data in textbox
        /// </summary>
        private void EEShow()
        {
            StringBuilder sb = new StringBuilder();
            for (int c = 0; c < 8; c++)
            {
                for (int l = 0; l < 8; l++)
                {
                    int n = c * 8 + l;
                    sb.Append( string.Format("{0}{1} ", _EEVal[n * 2 + 1].ToString("X2"), _EEVal[n * 2].ToString("X2")));
                }
                sb.AppendLine();
            }

            EEVals.Text = sb.ToString();
        }

        /// <summary>
        /// disables write to EEPROM
        /// </summary>
        private void disableWrite()
        {
            byte[] dgram = { (byte)'\'', (byte)'0', 0x00, 0xAA, 0x55 };
            if (_deviceEP != null)
                _udp.Send(dgram, dgram.Length, _deviceEP);
        }

        /// <summary>
        /// enables write to EEPROM
        /// </summary>
        private void enableWrite()
        {
            byte[] dgram = { (byte)'\'', (byte)'1', 0x00, 0xAA, 0x55 };
            if (_deviceEP != null)
                _udp.Send(dgram, dgram.Length, _deviceEP);
        }

        /// <summary>
        /// read specific EEPROM byte
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        private bool readEEByte(byte address)
        {
            byte[] dgram = { (byte)'\'', (byte)'R', 0x00, 0x00, 0x00 };
            dgram[2] = address;

            // send read request
            if (_deviceEP != null)
                _udp.Send(dgram, dgram.Length, _deviceEP);
            
            // give module time to respond
            System.Threading.Thread.Sleep(_EEReadDelay);

            return waitForData(1000);
        }

        /// <summary>
        /// write value to address
        /// </summary>
        /// <param name="address"></param>
        /// <param name="value"></param>
        private void writeEEbyte(byte address, int value)
        {
            byte[] dgram = { (byte)'\'', (byte)'W', 0x00, 0x00, 0x00 };
            dgram[2] = address;
            dgram[3] = (byte)(value / 256);
            dgram[4] = (byte)(value % 256);
            if (_deviceEP != null)
                _udp.Send(dgram, dgram.Length, _deviceEP);

            System.Threading.Thread.Sleep(100);
        }

        /// <summary>
        /// wait until data received or timeout has occured
        /// </summary>
        /// <param name="timeout"></param>
        private bool waitForData(double timeout)
        {
            tmrCheckAvailable.Stop();

            if (_udp == null)
                return false;

            bool rslt = false;

            DateTime now = DateTime.Now;
            while (!rslt && DateTime.Now < now.AddMilliseconds(timeout))
            {
                if (_udp != null && _udp.Available >=4)
                {
                    processUDPdata();
                    rslt = true;
                }
            }

            tmrCheckAvailable.Start();

            return rslt;
        }

        /// <summary>
        /// update PORT-A information
        /// </summary>
        /// <param name="sub_format"></param>
        /// <param name="read_value"></param>
        /// <param name="auto_read"></param>
        private void setPortA(int sub_format, byte read_value, bool auto_read)
        {
            switch (sub_format)
            {
                case 0:     // value
                    updateTestValueA(read_value);
                    lbTestValueAutoA.Text = auto_read ? "*" : "";
                    break;

                case 1:     // direction
                    updateTestDirectionA(read_value);
                    break;

                case 2:     // pullup
                    updateTestPullupA(read_value);
                    break;

                case 3:     // threshold
                    updateTestThresholdA(read_value);
                    break;

                case 4:     // schmitt
                    updateTestSchmittA(read_value);
                    break;
            }
        }

        /// <summary>
        /// update PORT-B information
        /// </summary>
        /// <param name="sub_format"></param>
        /// <param name="read_value"></param>
        /// <param name="auto_read"></param>
        private void setPortB(int sub_format, byte read_value, bool auto_read)
        {
            switch (sub_format)
            {
                case 0:     // value
                    updateTestValueB(read_value);
                    lbTestValueAutoB.Text = auto_read ? "*" : "";
                    break;

                case 1:     // direction
                    updateTestDirectionB(read_value);
                    break;

                case 2:     // pullup
                    updateTestPullupB(read_value);
                    break;

                case 3:     // threshold
                    updateTestThresholdB(read_value);
                    break;

                case 4:     // schmitt
                    updateTestSchmittB(read_value);
                    break;
            }
        }

        /// <summary>
        /// update PORT-C information
        /// </summary>
        /// <param name="sub_format"></param>
        /// <param name="read_value"></param>
        /// <param name="auto_read"></param>
        private void setPortC(int sub_format, byte read_value, bool auto_read)
        {
            switch (sub_format)
            {
                case 0:     // value
                    updateTestValueC(read_value);
                    lbTestValueAutoC.Text = auto_read ? "*" : "";
                    break;

                case 1:     // direction
                    updateTestDirectionC(read_value);
                    break;

                case 2:     // pullup
                    updateTestPullupC(read_value);
                    break;

                case 3:     // threshold
                    updateTestThresholdC(read_value);
                    break;

                case 4:     // schmitt
                    updateTestSchmittC(read_value);
                    break;
            }
        }

        /// <summary>
        /// update AutoLoad information
        /// </summary>
        /// <param name="byte_counter"></param>
        /// <param name="read_value"></param>
        /// <param name="liast_read"></param>
        private void setAutoLoad(int byte_counter, byte read_value, ref byte last_read)
        {
            switch (byte_counter)
            {
                // IP Address
                case 4:
                    tbAutoscanIP1.Text = read_value.ToString();
                    _EEVal[44] = read_value;
                    break;
                case 5:
                    tbAutoscanIP2.Text = read_value.ToString();
                    _EEVal[45] = read_value;
                    break;
                case 6:
                    tbAutoscanIP3.Text = read_value.ToString();
                    _EEVal[46] = read_value;
                    break;
                case 7:
                    tbAutoscanIP4.Text = read_value.ToString();
                    _EEVal[47] = read_value;
                    break;

                // MAC Address
                case 8:
                    tbAutoscanMAC1.Text = read_value.ToString("X2");
                    _EEVal[38] = read_value;
                    break;
                case 9:
                    tbAutoscanMAC2.Text = read_value.ToString("X2");
                    _EEVal[39] = read_value;
                    break;
                case 10:
                    tbAutoscanMAC3.Text = read_value.ToString("X2");
                    _EEVal[40] = read_value;
                    break;
                case 11:
                    tbAutoscanMAC4.Text = read_value.ToString("X2");
                    _EEVal[41] = read_value;
                    break;
                case 12:
                    tbAutoscanMAC5.Text = read_value.ToString("X2");
                    _EEVal[42] = read_value;
                    break;
                case 13:
                    tbAutoscanMAC6.Text = read_value.ToString("X2");
                    _EEVal[43] = read_value;
                    break;

                // Target port
                case 14:
                    last_read = read_value;
                    _EEVal[49] = read_value;
                    break;
                case 15:
                    tbAutoscanTargetPort.Text = (last_read * 256 + read_value).ToString();
                    _EEVal[48] = read_value;
                    break;

            }
        }

        /// <summary>
        /// process any UDP data waiting
        /// </summary>
        private void processUDPdata()
        {
            if (_udp != null)
            {
                // loop while there is data to collect
                while (_udp.Available > 0)
                {
                    IPEndPoint ep = new IPEndPoint(IPAddress.Parse("255.255.255.255"), 0);
                    byte[] data = _udp.Receive(ref ep);
                    string ipAddress = ep.Address.ToString();
                    string text = Encoding.UTF8.GetString(data);

                    if (data.Length < 4)
                    {
                        _EEReadDelay += 10;
                        debug.WriteLine("Short packet from {0}\r\nEEReadDelay increased to {1}\r\n", ipAddress, _EEReadDelay);
                    }
                    // check if this is an IO24 announcement
                    else if (text.Length > 4 && text.Substring(0, 4) == "IO24")
                    {
                        // only display information from modules
                        if (data.Length > 4)
                        {
                            if (!_ReadingEE)
                            {
                                // retrieve module information
                                string id = string.Format("{0} - MAC {1}:{2}:{3}:{4}:{5}:{6} V{7}.{8}",
                                    ipAddress,
                                    data[4].ToString("X2"), data[5].ToString("X2"), data[6].ToString("X2"), data[7].ToString("X2"), data[8].ToString("X2"), data[9].ToString("X2"),
                                    data[10], data[11]);

                                // add to list
                                ToolStripDropDownItem itm = new ToolStripMenuItem(id, null, new EventHandler(Module_Select));
                                itm.Tag = ep.Address.ToString();
                                tsddModules.DropDownItems.Add(itm);
                                tsddModules.Text = string.Format("{0} modules found", tsddModules.DropDownItems.Count);
                            }
                        }

                        // pause to allow time for other modules to respond
                        if (_udp.Available == 0)
                            System.Threading.Thread.Sleep(100);
                    }
                    else if (_deviceIP.ToString() != ipAddress)
                    {
                        // if message is not from the device we are addressing, 
                        // exit this loop (do nothing)
                        debug.WriteLine("Response from {0}", ipAddress);
                    }
                    else if (data.Length == 4)
                    {
                        // check if reading
                        if (data[0] == 'R' && !_EEWriteFlag)
                        {
                            /******************************************************************************************************
                             * READING EEPROM data from module
                             *****************************************************************************************************/
                            if (_EEAddress == data[1])
                            {
                                _EEVal[_EEAddress * 2 + 1] = data[2];
                                _EEVal[_EEAddress * 2] = data[3];
                                _EEReadSuccess = true;
                            }
                        }

                        // check if writing data to EEPROM
                        else if (_EEWriteFlag)
                        {
                            /******************************************************************************************************
                             * WRITING EEPROM data to module
                             *****************************************************************************************************/
                            if (_EEAddress == data[1] && _EEData == (data[2] * 256 + data[3]))
                            {
                                _EEAddress++;
                                _EEWriteSuccess = true;
                            }
                            else
                            {
                                _EEWriteSuccess = false;
                                disableWrite();
                                MessageBox.Show(string.Format("EEPROM Write Fail\nat Address {0}", _EEAddress),
                                    "EEPROM Write", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);

                                _EEWriteFlag = false;
                            }

                            if (_EEAddress < 63)
                            {
                                pbWriteStatus.Value = _EEAddress;
                                _EEData = _EEVal[_EEAddress * 2 + 1] * 256 + _EEVal[_EEAddress * 2];

                                writeEEbyte(_EEAddress, _EEData);
                                while (!readEEByte(_EEAddress));
                            }
                            else if (_EEWriteSuccess)
                            {
                                disableWrite();
                                pbWriteStatus.Value = pbWriteStatus.Maximum;

                                MessageBox.Show("EEPROM Write Successful",
                                    "EEPROM Write", MessageBoxButtons.OK, MessageBoxIcon.Information);

                                _EEWriteFlag = false;
                                pbWriteStatus.Value = 0;
                            }
                        }
                    }
                    else
                    {
                        /******************************************************************************************************
                         * EEPROM data
                         *****************************************************************************************************/
                        bool auto_read = false;
                        int byte_counter = 0;
                        int data_format = 0;
                        int sub_format = 0;
                        byte read_value = 0;
                        byte last_read = 0;
                        char ch = '\0';

                        // read through all data in packet
                        for (int n = 0; n < data.Length; n++)
                        {
                            read_value = data[n];
                            ch = (char)read_value;

                            switch (ch)
                            {
                                case '*':
                                    auto_read = true;
                                    break;

                                case '!':
                                    sub_format = 1;
                                    break;

                                case '@':
                                    sub_format = 2;
                                    break;

                                case '#':
                                    sub_format = 3;
                                    break;

                                case '$':
                                    sub_format = 4;
                                    break;

                                default:
                                    if (data_format == 0)
                                    {
                                        // new packet of data
                                        switch (ch)
                                        {
                                            case 'A':
                                                data_format = 1;
                                                break;

                                            case 'B':
                                                data_format = 2;
                                                break;

                                            case 'C':
                                                data_format = 3;
                                                break;

                                            case '%':
                                                data_format = 4;
                                                byte_counter = 1;
                                                break;
                                        }
                                    }
                                    else
                                    {
                                        // update information based on format
                                        switch (data_format)
                                        {
                                            case 1:
                                                setPortA(sub_format, read_value, auto_read);
                                                data_format = 0;
                                                sub_format = 0;
                                                break;

                                            case 2:
                                                setPortB(sub_format, read_value, auto_read);
                                                data_format = 0;
                                                sub_format = 0;
                                                break;

                                            case 3:
                                                setPortC(sub_format, read_value, auto_read);
                                                data_format = 0;
                                                sub_format = 0;
                                                break;

                                            case 4:
                                                if (_AutoLoadTarget)
                                                    setAutoLoad(byte_counter, read_value, ref last_read);

                                                // increment data pointer
                                                if (++byte_counter > 15)
                                                {
                                                    // we are done
                                                    _AutoLoadTarget = false;
                                                    data_format = 0;
                                                    sub_format = 0;
                                                }
                                                break;
                                        }
                                    }
                                    break;
                            }
                        }
                    }
                }
            }
        }
        #endregion

        #region Update TEST controls
        private void updateTestValueA(byte value)
        {
            bTestValueBitA0.Text = (value & 1) == 0 ? "0" : "1";
            bTestValueBitA1.Text = (value & 2) == 0 ? "0" : "1";
            bTestValueBitA2.Text = (value & 4) == 0 ? "0" : "1";
            bTestValueBitA3.Text = (value & 8) == 0 ? "0" : "1";
            bTestValueBitA4.Text = (value & 16) == 0 ? "0" : "1";
            bTestValueBitA5.Text = (value & 32) == 0 ? "0" : "1";
            bTestValueBitA6.Text = (value & 64) == 0 ? "0" : "1";
            bTestValueBitA7.Text = (value & 128) == 0 ? "0" : "1";
            tbTestValueHexA.Text = value.ToString("X2");
            tbTestValueDecimalA.Text = value.ToString();
        }

        private void updateTestValueB(byte value)
        {
            bTestValueBitB0.Text = (value & 1) == 0 ? "0" : "1";
            bTestValueBitB1.Text = (value & 2) == 0 ? "0" : "1";
            bTestValueBitB2.Text = (value & 4) == 0 ? "0" : "1";
            bTestValueBitB3.Text = (value & 8) == 0 ? "0" : "1";
            bTestValueBitB4.Text = (value & 16) == 0 ? "0" : "1";
            bTestValueBitB5.Text = (value & 32) == 0 ? "0" : "1";
            bTestValueBitB6.Text = (value & 64) == 0 ? "0" : "1";
            bTestValueBitB7.Text = (value & 128) == 0 ? "0" : "1";
            tbTestValueHexB.Text = value.ToString("X2");
            tbTestValueDecimalB.Text = value.ToString();
        }

        private void updateTestValueC(byte value)
        {
            bTestValueBitC0.Text = (value & 1) == 0 ? "0" : "1";
            bTestValueBitC1.Text = (value & 2) == 0 ? "0" : "1";
            bTestValueBitC2.Text = (value & 4) == 0 ? "0" : "1";
            bTestValueBitC3.Text = (value & 8) == 0 ? "0" : "1";
            bTestValueBitC4.Text = (value & 16) == 0 ? "0" : "1";
            bTestValueBitC5.Text = (value & 32) == 0 ? "0" : "1";
            bTestValueBitC6.Text = (value & 64) == 0 ? "0" : "1";
            bTestValueBitC7.Text = (value & 128) == 0 ? "0" : "1";
            tbTestValueHexC.Text = value.ToString("X2");
            tbTestValueDecimalC.Text = value.ToString();
        }

        private void updateTestDirectionA(byte value)
        {
            tbTestDirectionHexA.Text = value.ToString("X2");
            tbTestDirectionDecimalA.Text = value.ToString();
            bTestDirectionBitA7.Text = (value & 128) == 0 ? "O" : "I";
            bTestDirectionBitA6.Text = (value & 64) == 0 ? "O" : "I";
            bTestDirectionBitA5.Text = (value & 32) == 0 ? "O" : "I";
            bTestDirectionBitA4.Text = (value & 16) == 0 ? "O" : "I";
            bTestDirectionBitA3.Text = (value & 8) == 0 ? "O" : "I";
            bTestDirectionBitA2.Text = (value & 4) == 0 ? "O" : "I";
            bTestDirectionBitA1.Text = (value & 2) == 0 ? "O" : "I";
            bTestDirectionBitA0.Text = (value & 1) == 0 ? "O" : "I";
        }

        private void updateTestDirectionB(byte value)
        {
            tbTestDirectionHexB.Text = value.ToString("X2");
            tbTestDirectionDecimalB.Text = value.ToString();
            bTestDirectionBitB7.Text = (value & 128) == 0 ? "O" : "I";
            bTestDirectionBitB6.Text = (value & 64) == 0 ? "O" : "I";
            bTestDirectionBitB5.Text = (value & 32) == 0 ? "O" : "I";
            bTestDirectionBitB4.Text = (value & 16) == 0 ? "O" : "I";
            bTestDirectionBitB3.Text = (value & 8) == 0 ? "O" : "I";
            bTestDirectionBitB2.Text = (value & 4) == 0 ? "O" : "I";
            bTestDirectionBitB1.Text = (value & 2) == 0 ? "O" : "I";
            bTestDirectionBitB0.Text = (value & 1) == 0 ? "O" : "I";
        }

        private void updateTestDirectionC(byte value)
        {
            tbTestDirectionHexC.Text = value.ToString("X2");
            tbTestDirectionDecimalC.Text = value.ToString();
            bTestDirectionBitC7.Text = (value & 128) == 0 ? "O" : "I";
            bTestDirectionBitC6.Text = (value & 64) == 0 ? "O" : "I";
            bTestDirectionBitC5.Text = (value & 32) == 0 ? "O" : "I";
            bTestDirectionBitC4.Text = (value & 16) == 0 ? "O" : "I";
            bTestDirectionBitC3.Text = (value & 8) == 0 ? "O" : "I";
            bTestDirectionBitC2.Text = (value & 4) == 0 ? "O" : "I";
            bTestDirectionBitC1.Text = (value & 2) == 0 ? "O" : "I";
            bTestDirectionBitC0.Text = (value & 1) == 0 ? "O" : "I";
        }

        private void updateTestPullupA(byte value)
        {
            tbTestPullupHexA.Text = value.ToString("X2");
            tbTestPullupDecimalA.Text = value.ToString();
            bTestPullupBitA7.Text = (value & 128) == 0 ? "Y" : "N";
            bTestPullupBitA6.Text = (value & 64) == 0 ? "Y" : "N";
            bTestPullupBitA5.Text = (value & 32) == 0 ? "Y" : "N";
            bTestPullupBitA4.Text = (value & 16) == 0 ? "Y" : "N";
            bTestPullupBitA3.Text = (value & 8) == 0 ? "Y" : "N";
            bTestPullupBitA2.Text = (value & 4) == 0 ? "Y" : "N";
            bTestPullupBitA1.Text = (value & 2) == 0 ? "Y" : "N";
            bTestPullupBitA0.Text = (value & 1) == 0 ? "Y" : "N";
        }

        private void updateTestPullupB(byte value)
        {
            tbTestPullupHexB.Text = value.ToString("X2");
            tbTestPullupDecimalB.Text = value.ToString();
            bTestPullupBitB7.Text = (value & 128) == 0 ? "Y" : "N";
            bTestPullupBitB6.Text = (value & 64) == 0 ? "Y" : "N";
            bTestPullupBitB5.Text = (value & 32) == 0 ? "Y" : "N";
            bTestPullupBitB4.Text = (value & 16) == 0 ? "Y" : "N";
            bTestPullupBitB3.Text = (value & 8) == 0 ? "Y" : "N";
            bTestPullupBitB2.Text = (value & 4) == 0 ? "Y" : "N";
            bTestPullupBitB1.Text = (value & 2) == 0 ? "Y" : "N";
            bTestPullupBitB0.Text = (value & 1) == 0 ? "Y" : "N";
        }

        private void updateTestPullupC(byte value)
        {
            tbTestPullupHexC.Text = value.ToString("X2");
            tbTestPullupDecimalC.Text = value.ToString();
            bTestPullupBitC7.Text = (value & 128) == 0 ? "Y" : "N";
            bTestPullupBitC6.Text = (value & 64) == 0 ? "Y" : "N";
            bTestPullupBitC5.Text = (value & 32) == 0 ? "Y" : "N";
            bTestPullupBitC4.Text = (value & 16) == 0 ? "Y" : "N";
            bTestPullupBitC3.Text = (value & 8) == 0 ? "Y" : "N";
            bTestPullupBitC2.Text = (value & 4) == 0 ? "Y" : "N";
            bTestPullupBitC1.Text = (value & 2) == 0 ? "Y" : "N";
            bTestPullupBitC0.Text = (value & 1) == 0 ? "Y" : "N";
        }

        private void updateTestThresholdA(byte value)
        {
            tbTestThresholdHexA.Text = value.ToString("X2");
            tbTestThresholdDecimalA.Text = value.ToString();
            bTestThresholdBitA7.Text = (value & 128) == 0 ? "C" : "T";
            bTestThresholdBitA6.Text = (value & 64) == 0 ? "C" : "T";
            bTestThresholdBitA5.Text = (value & 32) == 0 ? "C" : "T";
            bTestThresholdBitA4.Text = (value & 16) == 0 ? "C" : "T";
            bTestThresholdBitA3.Text = (value & 8) == 0 ? "C" : "T";
            bTestThresholdBitA2.Text = (value & 4) == 0 ? "C" : "T";
            bTestThresholdBitA1.Text = (value & 2) == 0 ? "C" : "T";
            bTestThresholdBitA0.Text = (value & 1) == 0 ? "C" : "T";
        }

        private void updateTestThresholdB(byte value)
        {
            tbTestThresholdHexB.Text = value.ToString("X2");
            tbTestThresholdDecimalB.Text = value.ToString();
            bTestThresholdBitB7.Text = (value & 128) == 0 ? "C" : "T";
            bTestThresholdBitB6.Text = (value & 64) == 0 ? "C" : "T";
            bTestThresholdBitB5.Text = (value & 32) == 0 ? "C" : "T";
            bTestThresholdBitB4.Text = (value & 16) == 0 ? "C" : "T";
            bTestThresholdBitB3.Text = (value & 8) == 0 ? "C" : "T";
            bTestThresholdBitB2.Text = (value & 4) == 0 ? "C" : "T";
            bTestThresholdBitB1.Text = (value & 2) == 0 ? "C" : "T";
            bTestThresholdBitB0.Text = (value & 1) == 0 ? "C" : "T";
        }

        private void updateTestThresholdC(byte value)
        {
            tbTestThresholdHexC.Text = value.ToString("X2");
            tbTestThresholdDecimalC.Text = value.ToString();
            bTestThresholdBitC7.Text = (value & 128) == 0 ? "C" : "T";
            bTestThresholdBitC6.Text = (value & 64) == 0 ? "C" : "T";
            bTestThresholdBitC5.Text = (value & 32) == 0 ? "C" : "T";
            bTestThresholdBitC4.Text = (value & 16) == 0 ? "C" : "T";
            bTestThresholdBitC3.Text = (value & 8) == 0 ? "C" : "T";
            bTestThresholdBitC2.Text = (value & 4) == 0 ? "C" : "T";
            bTestThresholdBitC1.Text = (value & 2) == 0 ? "C" : "T";
            bTestThresholdBitC0.Text = (value & 1) == 0 ? "C" : "T";
        }

        private void updateTestSchmittA(byte value)
        {
            tbTestSchmittHexA.Text = value.ToString("X2");
            tbTestSchmittDecimalA.Text = value.ToString();
            bTestSchmittBitA7.Text = (value & 128) == 0 ? "S" : "N";
            bTestSchmittBitA6.Text = (value & 64) == 0 ? "S" : "N";
            bTestSchmittBitA5.Text = (value & 32) == 0 ? "S" : "N";
            bTestSchmittBitA4.Text = (value & 16) == 0 ? "S" : "N";
            bTestSchmittBitA3.Text = (value & 8) == 0 ? "S" : "N";
            bTestSchmittBitA2.Text = (value & 4) == 0 ? "S" : "N";
            bTestSchmittBitA1.Text = (value & 2) == 0 ? "S" : "N";
            bTestSchmittBitA0.Text = (value & 1) == 0 ? "S" : "N";
        }

        private void updateTestSchmittB(byte value)
        {
            tbTestSchmittHexB.Text = value.ToString("X2");
            tbTestSchmittDecimalB.Text = value.ToString();
            bTestSchmittBitB7.Text = (value & 128) == 0 ? "S" : "N";
            bTestSchmittBitB6.Text = (value & 64) == 0 ? "S" : "N";
            bTestSchmittBitB5.Text = (value & 32) == 0 ? "S" : "N";
            bTestSchmittBitB4.Text = (value & 16) == 0 ? "S" : "N";
            bTestSchmittBitB3.Text = (value & 8) == 0 ? "S" : "N";
            bTestSchmittBitB2.Text = (value & 4) == 0 ? "S" : "N";
            bTestSchmittBitB1.Text = (value & 2) == 0 ? "S" : "N";
            bTestSchmittBitB0.Text = (value & 1) == 0 ? "S" : "N";
        }

        private void updateTestSchmittC(byte value)
        {
            tbTestSchmittHexC.Text = value.ToString("X2");
            tbTestSchmittDecimalC.Text = value.ToString();
            bTestSchmittBitC7.Text = (value & 128) == 0 ? "S" : "N";
            bTestSchmittBitC6.Text = (value & 64) == 0 ? "S" : "N";
            bTestSchmittBitC5.Text = (value & 32) == 0 ? "S" : "N";
            bTestSchmittBitC4.Text = (value & 16) == 0 ? "S" : "N";
            bTestSchmittBitC3.Text = (value & 8) == 0 ? "S" : "N";
            bTestSchmittBitC2.Text = (value & 4) == 0 ? "S" : "N";
            bTestSchmittBitC1.Text = (value & 2) == 0 ? "S" : "N";
            bTestSchmittBitC0.Text = (value & 1) == 0 ? "S" : "N";
        }
        #endregion

        #region Update PROGRAM controls
        private void updateProgramValueA(byte value)
        {
            bProgramValueBitA0.Text = (value & 1) == 0 ? "0" : "1";
            bProgramValueBitA1.Text = (value & 2) == 0 ? "0" : "1";
            bProgramValueBitA2.Text = (value & 4) == 0 ? "0" : "1";
            bProgramValueBitA3.Text = (value & 8) == 0 ? "0" : "1";
            bProgramValueBitA4.Text = (value & 16) == 0 ? "0" : "1";
            bProgramValueBitA5.Text = (value & 32) == 0 ? "0" : "1";
            bProgramValueBitA6.Text = (value & 64) == 0 ? "0" : "1";
            bProgramValueBitA7.Text = (value & 128) == 0 ? "0" : "1";
            tbProgramValueHexA.Text = value.ToString("X2");
            tbProgramValueDecimalA.Text = value.ToString();

            _EEVal[17] = value;
            EEShow();
        }

        private void updateProgramValueB(byte value)
        {
            bProgramValueBitB0.Text = (value & 1) == 0 ? "0" : "1";
            bProgramValueBitB1.Text = (value & 2) == 0 ? "0" : "1";
            bProgramValueBitB2.Text = (value & 4) == 0 ? "0" : "1";
            bProgramValueBitB3.Text = (value & 8) == 0 ? "0" : "1";
            bProgramValueBitB4.Text = (value & 16) == 0 ? "0" : "1";
            bProgramValueBitB5.Text = (value & 32) == 0 ? "0" : "1";
            bProgramValueBitB6.Text = (value & 64) == 0 ? "0" : "1";
            bProgramValueBitB7.Text = (value & 128) == 0 ? "0" : "1";
            tbProgramValueHexB.Text = value.ToString("X2");
            tbProgramValueDecimalB.Text = value.ToString();

            _EEVal[22] = value;
            EEShow();
        }

        private void updateProgramValueC(byte value)
        {
            bProgramValueBitC0.Text = (value & 1) == 0 ? "0" : "1";
            bProgramValueBitC1.Text = (value & 2) == 0 ? "0" : "1";
            bProgramValueBitC2.Text = (value & 4) == 0 ? "0" : "1";
            bProgramValueBitC3.Text = (value & 8) == 0 ? "0" : "1";
            bProgramValueBitC4.Text = (value & 16) == 0 ? "0" : "1";
            bProgramValueBitC5.Text = (value & 32) == 0 ? "0" : "1";
            bProgramValueBitC6.Text = (value & 64) == 0 ? "0" : "1";
            bProgramValueBitC7.Text = (value & 128) == 0 ? "0" : "1";
            tbProgramValueHexC.Text = value.ToString("X2");
            tbProgramValueDecimalC.Text = value.ToString();

            _EEVal[27] = value;
            EEShow();
        }

        private void updateProgramDirectionA(byte value)
        {
            tbProgramDirectionHexA.Text = value.ToString("X2");
            tbProgramDirectionDecimalA.Text = value.ToString();
            bProgramDirectionBitA7.Text = (value & 128) == 0 ? "O" : "I";
            bProgramDirectionBitA6.Text = (value & 64) == 0 ? "O" : "I";
            bProgramDirectionBitA5.Text = (value & 32) == 0 ? "O" : "I";
            bProgramDirectionBitA4.Text = (value & 16) == 0 ? "O" : "I";
            bProgramDirectionBitA3.Text = (value & 8) == 0 ? "O" : "I";
            bProgramDirectionBitA2.Text = (value & 4) == 0 ? "O" : "I";
            bProgramDirectionBitA1.Text = (value & 2) == 0 ? "O" : "I";
            bProgramDirectionBitA0.Text = (value & 1) == 0 ? "O" : "I";

            _EEVal[16] = value;
            EEShow();
        }

        private void updateProgramDirectionB(byte value)
        {
            tbProgramDirectionHexB.Text = value.ToString("X2");
            tbProgramDirectionDecimalB.Text = value.ToString();
            bProgramDirectionBitB7.Text = (value & 128) == 0 ? "O" : "I";
            bProgramDirectionBitB6.Text = (value & 64) == 0 ? "O" : "I";
            bProgramDirectionBitB5.Text = (value & 32) == 0 ? "O" : "I";
            bProgramDirectionBitB4.Text = (value & 16) == 0 ? "O" : "I";
            bProgramDirectionBitB3.Text = (value & 8) == 0 ? "O" : "I";
            bProgramDirectionBitB2.Text = (value & 4) == 0 ? "O" : "I";
            bProgramDirectionBitB1.Text = (value & 2) == 0 ? "O" : "I";
            bProgramDirectionBitB0.Text = (value & 1) == 0 ? "O" : "I";

            _EEVal[21] = value;
            EEShow();
        }

        private void updateProgramDirectionC(byte value)
        {
            tbProgramDirectionHexC.Text = value.ToString("X2");
            tbProgramDirectionDecimalC.Text = value.ToString();
            bProgramDirectionBitC7.Text = (value & 128) == 0 ? "O" : "I";
            bProgramDirectionBitC6.Text = (value & 64) == 0 ? "O" : "I";
            bProgramDirectionBitC5.Text = (value & 32) == 0 ? "O" : "I";
            bProgramDirectionBitC4.Text = (value & 16) == 0 ? "O" : "I";
            bProgramDirectionBitC3.Text = (value & 8) == 0 ? "O" : "I";
            bProgramDirectionBitC2.Text = (value & 4) == 0 ? "O" : "I";
            bProgramDirectionBitC1.Text = (value & 2) == 0 ? "O" : "I";
            bProgramDirectionBitC0.Text = (value & 1) == 0 ? "O" : "I";

            _EEVal[26] = value;
            EEShow();
        }

        private void updateProgramPullupA(byte value)
        {
            tbProgramPullupHexA.Text = value.ToString("X2");
            tbProgramPullupDecimalA.Text = value.ToString();
            bProgramPullupBitA7.Text = (value & 128) == 0 ? "Y" : "N";
            bProgramPullupBitA6.Text = (value & 64) == 0 ? "Y" : "N";
            bProgramPullupBitA5.Text = (value & 32) == 0 ? "Y" : "N";
            bProgramPullupBitA4.Text = (value & 16) == 0 ? "Y" : "N";
            bProgramPullupBitA3.Text = (value & 8) == 0 ? "Y" : "N";
            bProgramPullupBitA2.Text = (value & 4) == 0 ? "Y" : "N";
            bProgramPullupBitA1.Text = (value & 2) == 0 ? "Y" : "N";
            bProgramPullupBitA0.Text = (value & 1) == 0 ? "Y" : "N";

            _EEVal[19] = value;
            EEShow();
        }

        private void updateProgramPullupB(byte value)
        {
            tbProgramPullupHexB.Text = value.ToString("X2");
            tbProgramPullupDecimalB.Text = value.ToString();
            bProgramPullupBitB7.Text = (value & 128) == 0 ? "Y" : "N";
            bProgramPullupBitB6.Text = (value & 64) == 0 ? "Y" : "N";
            bProgramPullupBitB5.Text = (value & 32) == 0 ? "Y" : "N";
            bProgramPullupBitB4.Text = (value & 16) == 0 ? "Y" : "N";
            bProgramPullupBitB3.Text = (value & 8) == 0 ? "Y" : "N";
            bProgramPullupBitB2.Text = (value & 4) == 0 ? "Y" : "N";
            bProgramPullupBitB1.Text = (value & 2) == 0 ? "Y" : "N";
            bProgramPullupBitB0.Text = (value & 1) == 0 ? "Y" : "N";

            _EEVal[24] = value;
            EEShow();
        }

        private void updateProgramPullupC(byte value)
        {
            tbProgramPullupHexC.Text = value.ToString("X2");
            tbProgramPullupDecimalC.Text = value.ToString();
            bProgramPullupBitC7.Text = (value & 128) == 0 ? "Y" : "N";
            bProgramPullupBitC6.Text = (value & 64) == 0 ? "Y" : "N";
            bProgramPullupBitC5.Text = (value & 32) == 0 ? "Y" : "N";
            bProgramPullupBitC4.Text = (value & 16) == 0 ? "Y" : "N";
            bProgramPullupBitC3.Text = (value & 8) == 0 ? "Y" : "N";
            bProgramPullupBitC2.Text = (value & 4) == 0 ? "Y" : "N";
            bProgramPullupBitC1.Text = (value & 2) == 0 ? "Y" : "N";
            bProgramPullupBitC0.Text = (value & 1) == 0 ? "Y" : "N";

            _EEVal[29] = value;
            EEShow();
        }

        private void updateProgramThresholdA(byte value)
        {
            tbProgramThresholdHexA.Text = value.ToString("X2");
            tbProgramThresholdDecimalA.Text = value.ToString();
            bProgramThresholdBitA7.Text = (value & 128) == 0 ? "C" : "T";
            bProgramThresholdBitA6.Text = (value & 64) == 0 ? "C" : "T";
            bProgramThresholdBitA5.Text = (value & 32) == 0 ? "C" : "T";
            bProgramThresholdBitA4.Text = (value & 16) == 0 ? "C" : "T";
            bProgramThresholdBitA3.Text = (value & 8) == 0 ? "C" : "T";
            bProgramThresholdBitA2.Text = (value & 4) == 0 ? "C" : "T";
            bProgramThresholdBitA1.Text = (value & 2) == 0 ? "C" : "T";
            bProgramThresholdBitA0.Text = (value & 1) == 0 ? "C" : "T";

            _EEVal[18] = value;
            EEShow();
        }

        private void updateProgramThresholdB(byte value)
        {
            tbProgramThresholdHexB.Text = value.ToString("X2");
            tbProgramThresholdDecimalB.Text = value.ToString();
            bProgramThresholdBitB7.Text = (value & 128) == 0 ? "C" : "T";
            bProgramThresholdBitB6.Text = (value & 64) == 0 ? "C" : "T";
            bProgramThresholdBitB5.Text = (value & 32) == 0 ? "C" : "T";
            bProgramThresholdBitB4.Text = (value & 16) == 0 ? "C" : "T";
            bProgramThresholdBitB3.Text = (value & 8) == 0 ? "C" : "T";
            bProgramThresholdBitB2.Text = (value & 4) == 0 ? "C" : "T";
            bProgramThresholdBitB1.Text = (value & 2) == 0 ? "C" : "T";
            bProgramThresholdBitB0.Text = (value & 1) == 0 ? "C" : "T";

            _EEVal[23] = value;
            EEShow();
        }

        private void updateProgramThresholdC(byte value)
        {
            tbProgramThresholdHexC.Text = value.ToString("X2");
            tbProgramThresholdDecimalC.Text = value.ToString();
            bProgramThresholdBitC7.Text = (value & 128) == 0 ? "C" : "T";
            bProgramThresholdBitC6.Text = (value & 64) == 0 ? "C" : "T";
            bProgramThresholdBitC5.Text = (value & 32) == 0 ? "C" : "T";
            bProgramThresholdBitC4.Text = (value & 16) == 0 ? "C" : "T";
            bProgramThresholdBitC3.Text = (value & 8) == 0 ? "C" : "T";
            bProgramThresholdBitC2.Text = (value & 4) == 0 ? "C" : "T";
            bProgramThresholdBitC1.Text = (value & 2) == 0 ? "C" : "T";
            bProgramThresholdBitC0.Text = (value & 1) == 0 ? "C" : "T";

            _EEVal[28] = value;
            EEShow();
        }

        private void updateProgramSchmittA(byte value)
        {
            tbProgramSchmittHexA.Text = value.ToString("X2");
            tbProgramSchmittDecimalA.Text = value.ToString();
            bProgramSchmittBitA7.Text = (value & 128) == 0 ? "S" : "N";
            bProgramSchmittBitA6.Text = (value & 64) == 0 ? "S" : "N";
            bProgramSchmittBitA5.Text = (value & 32) == 0 ? "S" : "N";
            bProgramSchmittBitA4.Text = (value & 16) == 0 ? "S" : "N";
            bProgramSchmittBitA3.Text = (value & 8) == 0 ? "S" : "N";
            bProgramSchmittBitA2.Text = (value & 4) == 0 ? "S" : "N";
            bProgramSchmittBitA1.Text = (value & 2) == 0 ? "S" : "N";
            bProgramSchmittBitA0.Text = (value & 1) == 0 ? "S" : "N";

            _EEVal[20] = value;
            EEShow();
        }

        private void updateProgramSchmittB(byte value)
        {
            tbProgramSchmittHexB.Text = value.ToString("X2");
            tbProgramSchmittDecimalB.Text = value.ToString();
            bProgramSchmittBitB7.Text = (value & 128) == 0 ? "S" : "N";
            bProgramSchmittBitB6.Text = (value & 64) == 0 ? "S" : "N";
            bProgramSchmittBitB5.Text = (value & 32) == 0 ? "S" : "N";
            bProgramSchmittBitB4.Text = (value & 16) == 0 ? "S" : "N";
            bProgramSchmittBitB3.Text = (value & 8) == 0 ? "S" : "N";
            bProgramSchmittBitB2.Text = (value & 4) == 0 ? "S" : "N";
            bProgramSchmittBitB1.Text = (value & 2) == 0 ? "S" : "N";
            bProgramSchmittBitB0.Text = (value & 1) == 0 ? "S" : "N";

            _EEVal[25] = value;
            EEShow();
        }

        private void updateProgramSchmittC(byte value)
        {
            tbProgramSchmittHexC.Text = value.ToString("X2");
            tbProgramSchmittDecimalC.Text = value.ToString();
            bProgramSchmittBitC7.Text = (value & 128) == 0 ? "S" : "N";
            bProgramSchmittBitC6.Text = (value & 64) == 0 ? "S" : "N";
            bProgramSchmittBitC5.Text = (value & 32) == 0 ? "S" : "N";
            bProgramSchmittBitC4.Text = (value & 16) == 0 ? "S" : "N";
            bProgramSchmittBitC3.Text = (value & 8) == 0 ? "S" : "N";
            bProgramSchmittBitC2.Text = (value & 4) == 0 ? "S" : "N";
            bProgramSchmittBitC1.Text = (value & 2) == 0 ? "S" : "N";
            bProgramSchmittBitC0.Text = (value & 1) == 0 ? "S" : "N";

            _EEVal[30] = value;
            EEShow();
        }
        #endregion

        #region Update AUTOSCAN controls
        private void updateAutoscanMaskA(byte value)
        {
            tbAutoscanMaskHexA.Text = value.ToString("X2");
            tbAutoscanMaskDecimalA.Text = value.ToString();
            bAutoscanMaskBitA7.Text = (value & 128) == 0 ? "0" : "1";
            bAutoscanMaskBitA6.Text = (value & 64) == 0 ? "0" : "1";
            bAutoscanMaskBitA5.Text = (value & 32) == 0 ? "0" : "1";
            bAutoscanMaskBitA4.Text = (value & 16) == 0 ? "0" : "1";
            bAutoscanMaskBitA3.Text = (value & 8) == 0 ? "0" : "1";
            bAutoscanMaskBitA2.Text = (value & 4) == 0 ? "0" : "1";
            bAutoscanMaskBitA1.Text = (value & 2) == 0 ? "0" : "1";
            bAutoscanMaskBitA0.Text = (value & 1) == 0 ? "0" : "1";

            _EEVal[32] = value;
            EEShow();
        }

        private void updateAutoscanMaskB(byte value)
        {
            tbAutoscanMaskHexB.Text = value.ToString("X2");
            tbAutoscanMaskDecimalB.Text = value.ToString();
            bAutoscanMaskBitB7.Text = (value & 128) == 0 ? "0" : "1";
            bAutoscanMaskBitB6.Text = (value & 64) == 0 ? "0" : "1";
            bAutoscanMaskBitB5.Text = (value & 32) == 0 ? "0" : "1";
            bAutoscanMaskBitB4.Text = (value & 16) == 0 ? "0" : "1";
            bAutoscanMaskBitB3.Text = (value & 8) == 0 ? "0" : "1";
            bAutoscanMaskBitB2.Text = (value & 4) == 0 ? "0" : "1";
            bAutoscanMaskBitB1.Text = (value & 2) == 0 ? "0" : "1";
            bAutoscanMaskBitB0.Text = (value & 1) == 0 ? "0" : "1";

            _EEVal[33] = value;
            EEShow();
        }

        private void updateAutoscanMaskC(byte value)
        {
            tbAutoscanMaskHexC.Text = value.ToString("X2");
            tbAutoscanMaskDecimalC.Text = value.ToString();
            bAutoscanMaskBitC7.Text = (value & 128) == 0 ? "0" : "1";
            bAutoscanMaskBitC6.Text = (value & 64) == 0 ? "0" : "1";
            bAutoscanMaskBitC5.Text = (value & 32) == 0 ? "0" : "1";
            bAutoscanMaskBitC4.Text = (value & 16) == 0 ? "0" : "1";
            bAutoscanMaskBitC3.Text = (value & 8) == 0 ? "0" : "1";
            bAutoscanMaskBitC2.Text = (value & 4) == 0 ? "0" : "1";
            bAutoscanMaskBitC1.Text = (value & 2) == 0 ? "0" : "1";
            bAutoscanMaskBitC0.Text = (value & 1) == 0 ? "0" : "1";

            _EEVal[34] = value;
            EEShow();
        }
        #endregion

        private void MainForm_Load(object sender, EventArgs e)
        {
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (_udp != null)
                _udp.Close();

            debug.AllowClose = true;
            tmrAutoRead.Stop();
        }

        /// <summary>
        /// exit application
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void File_Exit(object sender, EventArgs e)
        {
            this.Close();
        }

        /// <summary>
        /// open configuration file
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void File_Open(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Title = "Load Config file";
            ofd.Filter = "Config File (*.cfg)|*.cfg";
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    // read values from file
                    StreamReader sr = new StreamReader(ofd.FileName);
                    for (int n = 10; n < 128; n++)
                    {
                        string fr = sr.ReadLine();
                        _EEVal[n] = byte.Parse(fr);
                    }
                    sr.Close();

                    // update controls with information
                    loadControls();
                }
                catch(Exception ex)
                {
                    MessageBox.Show("Failed to load file\n\n" + ex.Message, "Load Config file", 
                        MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                }

            }
        }

        /// <summary>
        /// save data to file
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void File_Save(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Title = "Save Config file";
            sfd.Filter = "Config File (*.cfg)|*.cfg";
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    // read values from file
                    StreamWriter sr = new StreamWriter(sfd.FileName);
                    for (int n = 10; n < 128; n++)
                        sr.WriteLine(_EEVal[n]);
                    sr.Close();
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Failed to save file\n\n" + ex.Message, "Save Config file",
                        MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                }

            }
        }

        /// <summary>
        /// print config file
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void File_Print(object sender, EventArgs e)
        {
            ConfigPrint frm = new ConfigPrint();
            frm.Print(_EEVal);
        }

        /// <summary>
        /// print preview config file
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void File_PrintPreview(object sender, EventArgs e)
        {
            ConfigPrint frm = new ConfigPrint();
            frm.Preview(_EEVal);
        }

        /// <summary>
        /// scan for IO24 modules
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Module_Scan(object sender, EventArgs e)
        {
            tsbSearch.Enabled = false;

            // remove any existing drop-down items
            tsddModules.DropDownItems.Clear();
            tsddModules.Text = "No modules found";
            enableModuleItems(false);

            // broadcast "who's there"
            _udp.EnableBroadcast = true;
            _deviceIP = IPAddress.Parse("255.255.255.255");
            IPEndPoint ep = new IPEndPoint(_deviceIP, _devicePort);
            byte[] dgram = Encoding.UTF8.GetBytes("IO24");
            try
            {
                // send query packet
                _udp.Send(dgram, dgram.Length, ep);

                // wait for data
                waitForData(1000);
            }
            catch { }

            tsbSearch.Enabled = true;
        }

        /// <summary>
        /// write EPPROM data to module
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Module_Write(object sender, EventArgs e)
        {
            // move focus to the destination port box to 
            // ensure validation any data entry
            tbDestinationPort.Focus();

            // enable write to EEPROM
            enableWrite();

            _EEWriteFlag = true;
            _EEAddress = 5;
            _EEData = _EEVal[_EEAddress * 2 + 1] * 256 + _EEVal[_EEAddress * 2];

            // start write process
            writeEEbyte(_EEAddress, _EEData);

            // read byte back
            readEEByte(_EEAddress);
        }

        /// <summary>
        /// read information from module EEPROM
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Module_Read(object sender, EventArgs e)
        {
            if (_deviceEP == null)
                return;

            Cursor.Current = Cursors.WaitCursor;

            _ReadingEE = true;
            while (!readEEByte(0));

            // read all bytes and store data
            pbWriteStatus.Value = 0;
            _EEAddress = 255;
            for (byte n = 0; n < 64; n++)
            {
                // update progress bar & refresh screen
                pbWriteStatus.Value = n + 1;
                // flag the expected data address
                _EEAddress = n;
                // trigger read byte
                _EEReadSuccess = false;
                while (!_EEReadSuccess) 
                    readEEByte(n);
            }
            pbWriteStatus.Value = 0;

            // display information
            loadControls();

            _ReadingEE = false;

            Cursor.Current = Cursors.Default;
        }

        /// <summary>
        /// send reboot signal to target module
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Module_Reboot(object sender, EventArgs e)
        {
            byte[] dgram = { (byte)'\'', (byte)'@', 0x00, 0xAA, 0x55 };
            if (_deviceEP != null)
                _udp.Send(dgram, dgram.Length, _deviceEP);
        }

        /// <summary>
        /// set target device as selected module
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Module_Select(object sender, EventArgs e)
        {
            // update toolbar
            ToolStripMenuItem itm = (ToolStripMenuItem)sender;
            tsddModules.Text = itm.Text;
            enableModuleItems(true);

            // capture IP Address
            lDeviceIP.Text = (string)itm.Tag;
            _deviceIP = IPAddress.Parse(lDeviceIP.Text);
            _deviceEP = new IPEndPoint(_deviceIP, _devicePort);
            Application.DoEvents();

            // request EEPROM info
            Module_Read(sender, e);
        }

        /// <summary>
        /// display about box
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Help_About(object sender, EventArgs e)
        {
            AboutBox frm = new AboutBox();
            frm.ShowDialog();
        }

        /// <summary>
        /// set enabled status of Read/Write/Reboot options
        /// </summary>
        /// <param name="enabled"></param>
        private void enableModuleItems(bool enabled)
        {
            tsbRead.Enabled = enabled;
            tsbWrite.Enabled = enabled;
            tsbReset.Enabled = enabled;

            miReadFromModule.Enabled = enabled;
            miWriteToModule.Enabled = enabled;
            miRebootModule.Enabled = enabled;
        }

        /// <summary>
        /// display not-implemented message
        /// </summary>
        private void notImplemented()
        {
            MessageBox.Show("Not yet implemented");
        }

        /// <summary>
        /// destination port value changed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void tbDestinationPort_Validated(object sender, EventArgs e)
        {
            int port;
            if (int.TryParse(tbDestinationPort.Text, out port) && port != _devicePort)
            {
                // store new port
                _devicePort = port;

                // close current port
                _udp.Close();

                // open new port
                _udp = new UdpClient(_devicePort, AddressFamily.InterNetwork);
            }
        }

        /// <summary>
        /// enable / disable auto-read option
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void cbAutoRead_CheckedChanged(object sender, EventArgs e)
        {
            if (cbAutoRead.Checked)
                tmrAutoRead.Start();
            else
                tmrAutoRead.Stop();
        }

        /// <summary>
        /// auto-read values
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void tmrAutoRead_Tick(object sender, EventArgs e)
        {
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("abc");
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        /// <summary>
        /// assign the current IP address to the module
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void bUseCurrentIP_Click(object sender, EventArgs e)
        {
            string[] ip = lDeviceIP.Text.Split('.');
            tbFixedIP1.Text = ip[0];
            tbFixedIP2.Text = ip[1];
            tbFixedIP3.Text = ip[2];
            tbFixedIP4.Text = ip[3];
        }

        /// <summary>
        /// the state of the PowerUp Enabled flag changed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void PowerUpEnabled_Changed(object sender, EventArgs e)
        {
            if (!_SetPowerUp)
            {
                _SetPowerUp = true;
                bool value = ((CheckBox)sender).Checked;

                cbProgramValueEnable.Checked = value;
                cbProgramDirectionEnable.Checked = value;
                cbProgramPullupEnable.Checked = value;
                cbProgramThresholdEnable.Checked = value;
                cbProgramSchmittEnable.Checked = value;

                
                if (!_ReadingEE)
                {

                    if (value)
                        _EEVal[10] = (byte)(_EEVal[10] ^ 2);
                    else
                        _EEVal[10] = (byte)(_EEVal[10] | 2);
                }

                EEShow();

                
                
                _SetPowerUp = false;
            }
        }

        /// <summary>
        /// select all text if we tab into the box
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void TextBox_Enter(object sender, EventArgs e)
        {
            TextBox tbox = (TextBox)sender;
            tbox.SelectAll();
        }

        /// <summary>
        /// read the value of the ports
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void bReadPorts_Click(object sender, EventArgs e)
        {
            byte[] dgram = Encoding.UTF8.GetBytes(@"abc!a!b!c@a@b@c#a#b#c$a$b$c");
            if (_deviceEP != null)
                _udp.Send(dgram, dgram.Length, _deviceEP);
        }

        /// <summary>
        /// generic routine to validate value, update controls and send to module
        /// </summary>
        /// <param name="tb"></param>
        /// <param name="update"></param>
        /// <param name="moduleCode"></param>
        /// <returns></returns>
        private bool validateValue(string text, UpdateValues update, string moduleCode)
        {
            bool rslt = true;

            byte value;
            try
            {
                if (text.StartsWith("0x"))
                    value = Convert.ToByte(text, 16);
                else
                    value = byte.Parse(text);

                // update buttons
                if (update != null)
                    update(value);

                // send to module
                if (_deviceEP != null)
                {
                    byte[] dgram = Encoding.UTF8.GetBytes(moduleCode);
                    dgram[1] = value;
                    _udp.Send(dgram, dgram.Length, _deviceEP);
                }
            }
            catch 
            {
                rslt = false;
            }

            return rslt;
        }

        /********************************************************************************
        * TEST settings
        *********************************************************************************/
        #region TEST VALUE settings
        private void bTestValueBitA_Click(object sender, EventArgs e)
        {
            // get reference to textbox123
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "1") btn.Text = "0"; else btn.Text = "1";

            // update bit data
            byte value = 0;
            if (bTestValueBitA7.Text == "1") value += 128;
            if (bTestValueBitA6.Text == "1") value += 64;
            if (bTestValueBitA5.Text == "1") value += 32;
            if (bTestValueBitA4.Text == "1") value += 16;
            if (bTestValueBitA3.Text == "1") value += 8;
            if (bTestValueBitA2.Text == "1") value += 4;
            if (bTestValueBitA1.Text == "1") value += 2;
            if (bTestValueBitA0.Text == "1") value += 1;

            tbTestValueHexA.Text = value.ToString("X2");
            tbTestValueDecimalA.Text = value.ToString();

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("A ");
                dgram[1] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void bTestValueBitB_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "1") btn.Text = "0"; else btn.Text = "1";

            // update bit data
            byte value = 0;
            if (bTestValueBitB7.Text == "1") value += 128;
            if (bTestValueBitB6.Text == "1") value += 64;
            if (bTestValueBitB5.Text == "1") value += 32;
            if (bTestValueBitB4.Text == "1") value += 16;
            if (bTestValueBitB3.Text == "1") value += 8;
            if (bTestValueBitB2.Text == "1") value += 4;
            if (bTestValueBitB1.Text == "1") value += 2;
            if (bTestValueBitB0.Text == "1") value += 1;

            tbTestValueHexB.Text = value.ToString("X2");
            tbTestValueDecimalB.Text = value.ToString();

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("B ");
                dgram[1] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void bTestValueBitC_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "1") btn.Text = "0"; else btn.Text = "1";

            // update bit data
            byte value = 0;
            if (bTestValueBitC7.Text == "1") value += 128;
            if (bTestValueBitC6.Text == "1") value += 64;
            if (bTestValueBitC5.Text == "1") value += 32;
            if (bTestValueBitC4.Text == "1") value += 16;
            if (bTestValueBitC3.Text == "1") value += 8;
            if (bTestValueBitC2.Text == "1") value += 4;
            if (bTestValueBitC1.Text == "1") value += 2;
            if (bTestValueBitC0.Text == "1") value += 1;

            tbTestValueHexC.Text = value.ToString("X2");
            tbTestValueDecimalC.Text = value.ToString();

            _EEVal[27] = value;

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("C ");
                dgram[1] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void tbTestValueHexA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestValueHexA.Text, updateTestValueA, "A ");
        }

        private void tbTestValueDecimalA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestValueDecimalA.Text, updateTestValueA, "A ");
        }

        private void tbTestValueHexB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestValueHexB.Text, updateTestValueB, "B ");
        }

        private void tbTestValueDecimalB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestValueDecimalB.Text, updateTestValueB, "B ");
        }

        private void tbTestValueHexC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestValueHexC.Text, updateTestValueC, "C ");
        }

        private void tbTestValueDecimalC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestValueDecimalC.Text, updateTestValueC, "C ");
        }
        #endregion

        #region TEST DIRECTION settings
        private void bTestDirectionBitA_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "I") btn.Text = "O"; else btn.Text = "I";

            // update bit data
            byte value = 0;
            if (bTestDirectionBitA7.Text == "I") value += 128;
            if (bTestDirectionBitA6.Text == "I") value += 64;
            if (bTestDirectionBitA5.Text == "I") value += 32;
            if (bTestDirectionBitA4.Text == "I") value += 16;
            if (bTestDirectionBitA3.Text == "I") value += 8;
            if (bTestDirectionBitA2.Text == "I") value += 4;
            if (bTestDirectionBitA1.Text == "I") value += 2;
            if (bTestDirectionBitA0.Text == "I") value += 1;

            tbTestDirectionHexA.Text = value.ToString("X2");
            tbTestDirectionDecimalA.Text = value.ToString();

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("!A ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void bTestDirectionBitB_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "I") btn.Text = "O"; else btn.Text = "I";

            // update bit data
            byte value = 0;
            if (bTestDirectionBitB7.Text == "I") value += 128;
            if (bTestDirectionBitB6.Text == "I") value += 64;
            if (bTestDirectionBitB5.Text == "I") value += 32;
            if (bTestDirectionBitB4.Text == "I") value += 16;
            if (bTestDirectionBitB3.Text == "I") value += 8;
            if (bTestDirectionBitB2.Text == "I") value += 4;
            if (bTestDirectionBitB1.Text == "I") value += 2;
            if (bTestDirectionBitB0.Text == "I") value += 1;

            tbTestDirectionHexB.Text = value.ToString("X2");
            tbTestDirectionDecimalB.Text = value.ToString();

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("!B ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void bTestDirectionBitC_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "I") btn.Text = "O"; else btn.Text = "I";

            // update bit data
            byte value = 0;
            if (bTestDirectionBitC7.Text == "I") value += 128;
            if (bTestDirectionBitC6.Text == "I") value += 64;
            if (bTestDirectionBitC5.Text == "I") value += 32;
            if (bTestDirectionBitC4.Text == "I") value += 16;
            if (bTestDirectionBitC3.Text == "I") value += 8;
            if (bTestDirectionBitC2.Text == "I") value += 4;
            if (bTestDirectionBitC1.Text == "I") value += 2;
            if (bTestDirectionBitC0.Text == "I") value += 1;

            tbTestDirectionHexC.Text = value.ToString("X2");
            tbTestDirectionDecimalC.Text = value.ToString();

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("!C ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void tbTestDirectionHexA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestDirectionHexA.Text, updateTestDirectionA, "!A ");
        }

        private void tbTestDirectionDecimalA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestDirectionDecimalA.Text, updateTestDirectionA, "!A ");
        }

        private void tbTestDirectionHexB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestDirectionHexB.Text, updateTestDirectionB, "!B ");
        }

        private void tbTestDirectionDecimalB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestDirectionDecimalB.Text, updateTestDirectionB, "!B ");
        }

        private void tbTestDirectionHexC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestDirectionHexC.Text, updateTestDirectionC, "!C ");
        }

        private void tbTestDirectionDecimalC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestDirectionDecimalC.Text, updateTestDirectionC, "!C ");
        }
        #endregion

        #region TEST PULL-UP settings
        private void bTestPullupBitA_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "Y") btn.Text = "N"; else btn.Text = "Y";

            // update bit data
            byte value = 0;
            if (bTestPullupBitA7.Text == "N") value += 128;
            if (bTestPullupBitA6.Text == "N") value += 64;
            if (bTestPullupBitA5.Text == "N") value += 32;
            if (bTestPullupBitA4.Text == "N") value += 16;
            if (bTestPullupBitA3.Text == "N") value += 8;
            if (bTestPullupBitA2.Text == "N") value += 4;
            if (bTestPullupBitA1.Text == "N") value += 2;
            if (bTestPullupBitA0.Text == "N") value += 1;

            tbTestPullupHexA.Text = value.ToString("X2");
            tbTestPullupDecimalA.Text = value.ToString();

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("@A ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void bTestPullupBitB_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "Y") btn.Text = "N"; else btn.Text = "Y";

            // update bit data
            byte value = 0;
            if (bTestPullupBitB7.Text == "N") value += 128;
            if (bTestPullupBitB6.Text == "N") value += 64;
            if (bTestPullupBitB5.Text == "N") value += 32;
            if (bTestPullupBitB4.Text == "N") value += 16;
            if (bTestPullupBitB3.Text == "N") value += 8;
            if (bTestPullupBitB2.Text == "N") value += 4;
            if (bTestPullupBitB1.Text == "N") value += 2;
            if (bTestPullupBitB0.Text == "N") value += 1;

            tbTestPullupHexB.Text = value.ToString("X2");
            tbTestPullupDecimalB.Text = value.ToString();

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("@B ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void bTestPullupBitC_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "Y") btn.Text = "N"; else btn.Text = "Y";

            // update bit data
            byte value = 0;
            if (bTestPullupBitC7.Text == "N") value += 128;
            if (bTestPullupBitC6.Text == "N") value += 64;
            if (bTestPullupBitC5.Text == "N") value += 32;
            if (bTestPullupBitC4.Text == "N") value += 16;
            if (bTestPullupBitC3.Text == "N") value += 8;
            if (bTestPullupBitC2.Text == "N") value += 4;
            if (bTestPullupBitC1.Text == "N") value += 2;
            if (bTestPullupBitC0.Text == "N") value += 1;

            tbTestPullupHexC.Text = value.ToString("X2");
            tbTestPullupDecimalC.Text = value.ToString();

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("@C ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void tbTestPullupHexA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestPullupHexA.Text, updateTestPullupA, "@A ");
        }

        private void tbTestPullupDecimalA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestPullupDecimalA.Text, updateTestPullupA, "@A ");
        }

        private void tbTestPullupHexB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestPullupHexB.Text, updateTestPullupB, "@B ");
        }

        private void tbTestPullupDecimalB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestPullupDecimalB.Text, updateTestPullupB, "@B ");
        }

        private void tbTestPullupHexC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestPullupHexC.Text, updateTestPullupC, "@C ");
        }

        private void tbTestPullupDecimalC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestPullupDecimalC.Text, updateTestPullupC, "@C ");
        }
        #endregion

        #region TEST THRESHOLD settings
        private void bTestThresholdBitA_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "T") btn.Text = "C"; else btn.Text = "T";

            // update bit data
            byte value = 0;
            if (bTestThresholdBitA7.Text == "T") value += 128;
            if (bTestThresholdBitA6.Text == "T") value += 64;
            if (bTestThresholdBitA5.Text == "T") value += 32;
            if (bTestThresholdBitA4.Text == "T") value += 16;
            if (bTestThresholdBitA3.Text == "T") value += 8;
            if (bTestThresholdBitA2.Text == "T") value += 4;
            if (bTestThresholdBitA1.Text == "T") value += 2;
            if (bTestThresholdBitA0.Text == "T") value += 1;

            tbTestThresholdHexA.Text = value.ToString("X2");
            tbTestThresholdDecimalA.Text = value.ToString();

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("#A ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void bTestThresholdBitB_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "T") btn.Text = "C"; else btn.Text = "T";

            // update bit data
            byte value = 0;
            if (bTestThresholdBitB7.Text == "T") value += 128;
            if (bTestThresholdBitB6.Text == "T") value += 64;
            if (bTestThresholdBitB5.Text == "T") value += 32;
            if (bTestThresholdBitB4.Text == "T") value += 16;
            if (bTestThresholdBitB3.Text == "T") value += 8;
            if (bTestThresholdBitB2.Text == "T") value += 4;
            if (bTestThresholdBitB1.Text == "T") value += 2;
            if (bTestThresholdBitB0.Text == "T") value += 1;

            tbTestThresholdHexB.Text = value.ToString("X2");
            tbTestThresholdDecimalB.Text = value.ToString();

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("#B ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void bTestThresholdBitC_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "T") btn.Text = "C"; else btn.Text = "T";

            // update bit data
            byte value = 0;
            if (bTestThresholdBitC7.Text == "T") value += 128;
            if (bTestThresholdBitC6.Text == "T") value += 64;
            if (bTestThresholdBitC5.Text == "T") value += 32;
            if (bTestThresholdBitC4.Text == "T") value += 16;
            if (bTestThresholdBitC3.Text == "T") value += 8;
            if (bTestThresholdBitC2.Text == "T") value += 4;
            if (bTestThresholdBitC1.Text == "T") value += 2;
            if (bTestThresholdBitC0.Text == "T") value += 1;

            tbTestThresholdHexC.Text = value.ToString("X2");
            tbTestThresholdDecimalC.Text = value.ToString();

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("#C ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void tbTestThresholdHexA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestThresholdHexA.Text, updateTestThresholdA, "#A ");
        }

        private void tbTestThresholdDecimalA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestThresholdDecimalA.Text, updateTestThresholdA, "#A ");
        }

        private void tbTestThresholdHexB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestThresholdHexB.Text, updateTestThresholdB, "#B ");
        }

        private void tbTestThresholdDecimalB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestThresholdDecimalB.Text, updateTestThresholdB, "#B ");
        }

        private void tbTestThresholdHexC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestThresholdHexC.Text, updateTestThresholdC, "#C ");
        }

        private void tbTestThresholdDecimalC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestThresholdDecimalC.Text, updateTestThresholdC, "#C ");
        }
        #endregion

        #region TEST SCHMITT settings
        private void bTestSchmittBitA_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "N") btn.Text = "S"; else btn.Text = "N";

            // update bit data
            byte value = 0;
            if (bTestSchmittBitA7.Text == "N") value += 128;
            if (bTestSchmittBitA6.Text == "N") value += 64;
            if (bTestSchmittBitA5.Text == "N") value += 32;
            if (bTestSchmittBitA4.Text == "N") value += 16;
            if (bTestSchmittBitA3.Text == "N") value += 8;
            if (bTestSchmittBitA2.Text == "N") value += 4;
            if (bTestSchmittBitA1.Text == "N") value += 2;
            if (bTestSchmittBitA0.Text == "N") value += 1;

            tbTestSchmittHexA.Text = value.ToString("X2");
            tbTestSchmittDecimalA.Text = value.ToString();

            _EEVal[20] = value;

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("$A ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void bTestSchmittBitB_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "N") btn.Text = "S"; else btn.Text = "N";

            // update bit data
            byte value = 0;
            if (bTestSchmittBitB7.Text == "N") value += 128;
            if (bTestSchmittBitB6.Text == "N") value += 64;
            if (bTestSchmittBitB5.Text == "N") value += 32;
            if (bTestSchmittBitB4.Text == "N") value += 16;
            if (bTestSchmittBitB3.Text == "N") value += 8;
            if (bTestSchmittBitB2.Text == "N") value += 4;
            if (bTestSchmittBitB1.Text == "N") value += 2;
            if (bTestSchmittBitB0.Text == "N") value += 1;

            tbTestSchmittHexB.Text = value.ToString("X2");
            tbTestSchmittDecimalB.Text = value.ToString();

            _EEVal[25] = value;

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("$B ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void bTestSchmittBitC_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "N") btn.Text = "S"; else btn.Text = "N";

            // update bit data
            byte value = 0;
            if (bTestSchmittBitC7.Text == "N") value += 128;
            if (bTestSchmittBitC6.Text == "N") value += 64;
            if (bTestSchmittBitC5.Text == "N") value += 32;
            if (bTestSchmittBitC4.Text == "N") value += 16;
            if (bTestSchmittBitC3.Text == "N") value += 8;
            if (bTestSchmittBitC2.Text == "N") value += 4;
            if (bTestSchmittBitC1.Text == "N") value += 2;
            if (bTestSchmittBitC0.Text == "N") value += 1;

            tbTestSchmittHexC.Text = value.ToString("X2");
            tbTestSchmittDecimalC.Text = value.ToString();

            _EEVal[30] = value;

            // send to module
            if (_deviceEP != null)
            {
                byte[] dgram = Encoding.UTF8.GetBytes("$C ");
                dgram[2] = value;
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void tbTestSchmittHexA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestSchmittHexA.Text, updateTestSchmittA, "$A ");
        }

        private void tbTestSchmittDecimalA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestSchmittDecimalA.Text, updateTestSchmittA, "$A ");
        }

        private void tbTestSchmittHexB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestSchmittHexB.Text, updateTestSchmittB, "$B ");
        }

        private void tbTestSchmittDecimalB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestSchmittDecimalB.Text, updateTestSchmittB, "$B ");
        }

        private void tbTestSchmittHexC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbTestSchmittHexC.Text, updateTestSchmittC, "$C ");
        }

        private void tbTestSchmittDecimalC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbTestSchmittDecimalC.Text, updateTestSchmittC, "$C ");
        }
        #endregion

        /*******************************************************************************
         * PROGRAM settings
         *******************************************************************************/
        #region PROGRAM VALUE settings
        private void bProgramValueBitA_Click(object sender, EventArgs e)
        {
            // get reference to textbox1234
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "1") btn.Text = "0"; else btn.Text = "1";

            // update bit data
            byte value = 0;
            if (bProgramValueBitA7.Text == "1") value += 128;
            if (bProgramValueBitA6.Text == "1") value += 64;
            if (bProgramValueBitA5.Text == "1") value += 32;
            if (bProgramValueBitA4.Text == "1") value += 16;
            if (bProgramValueBitA3.Text == "1") value += 8;
            if (bProgramValueBitA2.Text == "1") value += 4;
            if (bProgramValueBitA1.Text == "1") value += 2;
            if (bProgramValueBitA0.Text == "1") value += 1;

            tbProgramValueHexA.Text = value.ToString("X2");
            tbProgramValueDecimalA.Text = value.ToString();
            _EEVal[17] = value;

        }

        private void bProgramValueBitB_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "1") btn.Text = "0"; else btn.Text = "1";

            // update bit data
            byte value = 0;
            if (bProgramValueBitB7.Text == "1") value += 128;
            if (bProgramValueBitB6.Text == "1") value += 64;
            if (bProgramValueBitB5.Text == "1") value += 32;
            if (bProgramValueBitB4.Text == "1") value += 16;
            if (bProgramValueBitB3.Text == "1") value += 8;
            if (bProgramValueBitB2.Text == "1") value += 4;
            if (bProgramValueBitB1.Text == "1") value += 2;
            if (bProgramValueBitB0.Text == "1") value += 1;

            tbProgramValueHexB.Text = value.ToString("X2");
            tbProgramValueDecimalB.Text = value.ToString();

            _EEVal[22] = value;
        }

        private void bProgramValueBitC_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "1") btn.Text = "0"; else btn.Text = "1";

            // update bit data
            byte value = 0;
            if (bProgramValueBitC7.Text == "1") value += 128;
            if (bProgramValueBitC6.Text == "1") value += 64;
            if (bProgramValueBitC5.Text == "1") value += 32;
            if (bProgramValueBitC4.Text == "1") value += 16;
            if (bProgramValueBitC3.Text == "1") value += 8;
            if (bProgramValueBitC2.Text == "1") value += 4;
            if (bProgramValueBitC1.Text == "1") value += 2;
            if (bProgramValueBitC0.Text == "1") value += 1;

            tbProgramValueHexC.Text = value.ToString("X2");
            tbProgramValueDecimalC.Text = value.ToString();

            _EEVal[27] = value;
        }

        private void tbProgramValueHexA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramValueHexA.Text, updateProgramValueA, null);
        }

        private void tbProgramValueDecimalA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramValueDecimalA.Text, updateProgramValueA, null);
        }

        private void tbProgramValueHexB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramValueHexB.Text, updateProgramValueB, null);
        }

        private void tbProgramValueDecimalB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramValueDecimalB.Text, updateProgramValueB, null);
        }

        private void tbProgramValueHexC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramValueHexC.Text, updateProgramValueC, null);
        }

        private void tbProgramValueDecimalC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramValueDecimalC.Text, updateProgramValueC, null);
        }
        #endregion

        #region PROGRAM DIRECTION settings
        private void bProgramDirectionBitA_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "I") btn.Text = "O"; else btn.Text = "I";

            // update bit data
            byte value = 0;
            if (bProgramDirectionBitA7.Text == "I") value += 128;
            if (bProgramDirectionBitA6.Text == "I") value += 64;
            if (bProgramDirectionBitA5.Text == "I") value += 32;
            if (bProgramDirectionBitA4.Text == "I") value += 16;
            if (bProgramDirectionBitA3.Text == "I") value += 8;
            if (bProgramDirectionBitA2.Text == "I") value += 4;
            if (bProgramDirectionBitA1.Text == "I") value += 2;
            if (bProgramDirectionBitA0.Text == "I") value += 1;

            tbProgramDirectionHexA.Text = value.ToString("X2");
            tbProgramDirectionDecimalA.Text = value.ToString();

            _EEVal[16] = value;
        }

        private void bProgramDirectionBitB_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "I") btn.Text = "O"; else btn.Text = "I";

            // update bit data
            byte value = 0;
            if (bProgramDirectionBitB7.Text == "I") value += 128;
            if (bProgramDirectionBitB6.Text == "I") value += 64;
            if (bProgramDirectionBitB5.Text == "I") value += 32;
            if (bProgramDirectionBitB4.Text == "I") value += 16;
            if (bProgramDirectionBitB3.Text == "I") value += 8;
            if (bProgramDirectionBitB2.Text == "I") value += 4;
            if (bProgramDirectionBitB1.Text == "I") value += 2;
            if (bProgramDirectionBitB0.Text == "I") value += 1;

            tbProgramDirectionHexB.Text = value.ToString("X2");
            tbProgramDirectionDecimalB.Text = value.ToString();

            _EEVal[21] = value;
        }

        private void bProgramDirectionBitC_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "I") btn.Text = "O"; else btn.Text = "I";

            // update bit data
            byte value = 0;
            if (bProgramDirectionBitC7.Text == "I") value += 128;
            if (bProgramDirectionBitC6.Text == "I") value += 64;
            if (bProgramDirectionBitC5.Text == "I") value += 32;
            if (bProgramDirectionBitC4.Text == "I") value += 16;
            if (bProgramDirectionBitC3.Text == "I") value += 8;
            if (bProgramDirectionBitC2.Text == "I") value += 4;
            if (bProgramDirectionBitC1.Text == "I") value += 2;
            if (bProgramDirectionBitC0.Text == "I") value += 1;

            tbProgramDirectionHexC.Text = value.ToString("X2");
            tbProgramDirectionDecimalC.Text = value.ToString();

            _EEVal[26] = value;
        }

        private void tbProgramDirectionHexA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramDirectionHexA.Text, updateProgramDirectionA, null);
        }

        private void tbProgramDirectionDecimalA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramDirectionDecimalA.Text, updateProgramDirectionA, null);
        }

        private void tbProgramDirectionHexB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramDirectionHexB.Text, updateProgramDirectionB, null);
        }

        private void tbProgramDirectionDecimalB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramDirectionDecimalB.Text, updateProgramDirectionB, null);
        }

        private void tbProgramDirectionHexC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramDirectionHexC.Text, updateProgramDirectionC, null);
        }

        private void tbProgramDirectionDecimalC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramDirectionDecimalC.Text, updateProgramDirectionC, null);
        }
        #endregion

        #region PROGRAM PULL-UP settings
        private void bProgramPullupBitA_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "N") btn.Text = "Y"; else btn.Text = "N";

            // update bit data
            byte value = 0;
            if (bProgramPullupBitA7.Text == "N") value += 128;
            if (bProgramPullupBitA6.Text == "N") value += 64;
            if (bProgramPullupBitA5.Text == "N") value += 32;
            if (bProgramPullupBitA4.Text == "N") value += 16;
            if (bProgramPullupBitA3.Text == "N") value += 8;
            if (bProgramPullupBitA2.Text == "N") value += 4;
            if (bProgramPullupBitA1.Text == "N") value += 2;
            if (bProgramPullupBitA0.Text == "N") value += 1;

            tbProgramPullupHexA.Text = value.ToString("X2");
            tbProgramPullupDecimalA.Text = value.ToString();

            _EEVal[19] = value;
        }

        private void bProgramPullupBitB_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "N") btn.Text = "Y"; else btn.Text = "N";

            // update bit data
            byte value = 0;
            if (bProgramPullupBitB7.Text == "N") value += 128;
            if (bProgramPullupBitB6.Text == "N") value += 64;
            if (bProgramPullupBitB5.Text == "N") value += 32;
            if (bProgramPullupBitB4.Text == "N") value += 16;
            if (bProgramPullupBitB3.Text == "N") value += 8;
            if (bProgramPullupBitB2.Text == "N") value += 4;
            if (bProgramPullupBitB1.Text == "N") value += 2;
            if (bProgramPullupBitB0.Text == "N") value += 1;

            tbProgramPullupHexB.Text = value.ToString("X2");
            tbProgramPullupDecimalB.Text = value.ToString();

            _EEVal[24] = value;
        }

        private void bProgramPullupBitC_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "N") btn.Text = "Y"; else btn.Text = "N";

            // update bit data
            byte value = 0;
            if (bProgramPullupBitC7.Text == "N") value += 128;
            if (bProgramPullupBitC6.Text == "N") value += 64;
            if (bProgramPullupBitC5.Text == "N") value += 32;
            if (bProgramPullupBitC4.Text == "N") value += 16;
            if (bProgramPullupBitC3.Text == "N") value += 8;
            if (bProgramPullupBitC2.Text == "N") value += 4;
            if (bProgramPullupBitC1.Text == "N") value += 2;
            if (bProgramPullupBitC0.Text == "N") value += 1;

            tbProgramPullupHexC.Text = value.ToString("X2");
            tbProgramPullupDecimalC.Text = value.ToString();

            _EEVal[29] = value;
        }

        private void tbProgramPullupHexA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramPullupHexA.Text, updateProgramPullupA, null);
        }

        private void tbProgramPullupDecimalA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramPullupDecimalA.Text, updateProgramPullupA, null);
        }

        private void tbProgramPullupHexB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramPullupHexB.Text, updateProgramPullupB, null);
        }

        private void tbProgramPullupDecimalB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramPullupDecimalB.Text, updateProgramPullupB, null);
        }

        private void tbProgramPullupHexC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramPullupHexC.Text, updateProgramPullupC, null);
        }

        private void tbProgramPullupDecimalC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramPullupDecimalC.Text, updateProgramPullupC, null);
        }
        #endregion

        #region PROGRAM THRESHOLD settings
        private void bProgramThresholdBitA_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "T") btn.Text = "C"; else btn.Text = "T";

            // update bit data
            byte value = 0;
            if (bProgramThresholdBitA7.Text == "T") value += 128;
            if (bProgramThresholdBitA6.Text == "T") value += 64;
            if (bProgramThresholdBitA5.Text == "T") value += 32;
            if (bProgramThresholdBitA4.Text == "T") value += 16;
            if (bProgramThresholdBitA3.Text == "T") value += 8;
            if (bProgramThresholdBitA2.Text == "T") value += 4;
            if (bProgramThresholdBitA1.Text == "T") value += 2;
            if (bProgramThresholdBitA0.Text == "T") value += 1;

            tbProgramThresholdHexA.Text = value.ToString("X2");
            tbProgramThresholdDecimalA.Text = value.ToString();

            _EEVal[18] = value;
        }

        private void bProgramThresholdBitB_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "T") btn.Text = "C"; else btn.Text = "T";

            // update bit data
            byte value = 0;
            if (bProgramThresholdBitB7.Text == "T") value += 128;
            if (bProgramThresholdBitB6.Text == "T") value += 64;
            if (bProgramThresholdBitB5.Text == "T") value += 32;
            if (bProgramThresholdBitB4.Text == "T") value += 16;
            if (bProgramThresholdBitB3.Text == "T") value += 8;
            if (bProgramThresholdBitB2.Text == "T") value += 4;
            if (bProgramThresholdBitB1.Text == "T") value += 2;
            if (bProgramThresholdBitB0.Text == "T") value += 1;

            tbProgramThresholdHexB.Text = value.ToString("X2");
            tbProgramThresholdDecimalB.Text = value.ToString();

            _EEVal[23] = value;
        }

        private void bProgramThresholdBitC_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "T") btn.Text = "C"; else btn.Text = "T";

            // update bit data
            byte value = 0;
            if (bProgramThresholdBitC7.Text == "T") value += 128;
            if (bProgramThresholdBitC6.Text == "T") value += 64;
            if (bProgramThresholdBitC5.Text == "T") value += 32;
            if (bProgramThresholdBitC4.Text == "T") value += 16;
            if (bProgramThresholdBitC3.Text == "T") value += 8;
            if (bProgramThresholdBitC2.Text == "T") value += 4;
            if (bProgramThresholdBitC1.Text == "T") value += 2;
            if (bProgramThresholdBitC0.Text == "T") value += 1;

            tbProgramThresholdHexC.Text = value.ToString("X2");
            tbProgramThresholdDecimalC.Text = value.ToString();

            _EEVal[28] = value;
        }

        private void tbProgramThresholdHexA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramThresholdHexA.Text, updateProgramThresholdA, null);
        }

        private void tbProgramThresholdDecimalA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramThresholdDecimalA.Text, updateProgramThresholdA, null);
        }

        private void tbProgramThresholdHexB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramThresholdHexB.Text, updateProgramThresholdB, null);
        }

        private void tbProgramThresholdDecimalB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramThresholdDecimalB.Text, updateProgramThresholdB, null);
        }

        private void tbProgramThresholdHexC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramThresholdHexC.Text, updateProgramThresholdC, null);
        }

        private void tbProgramThresholdDecimalC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramThresholdDecimalC.Text, updateProgramThresholdC, null);
        }
        #endregion

        #region PROGRAM SCHMITT settings
        private void bProgramSchmittBitA_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "N") btn.Text = "S"; else btn.Text = "N";

            // update bit data
            byte value = 0;
            if (bProgramSchmittBitA7.Text == "N") value += 128;
            if (bProgramSchmittBitA6.Text == "N") value += 64;
            if (bProgramSchmittBitA5.Text == "N") value += 32;
            if (bProgramSchmittBitA4.Text == "N") value += 16;
            if (bProgramSchmittBitA3.Text == "N") value += 8;
            if (bProgramSchmittBitA2.Text == "N") value += 4;
            if (bProgramSchmittBitA1.Text == "N") value += 2;
            if (bProgramSchmittBitA0.Text == "N") value += 1;

            tbProgramSchmittHexA.Text = value.ToString("X2");
            tbProgramSchmittDecimalA.Text = value.ToString();
        }

        private void bProgramSchmittBitB_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "N") btn.Text = "S"; else btn.Text = "N";

            // update bit data
            byte value = 0;
            if (bProgramSchmittBitB7.Text == "N") value += 128;
            if (bProgramSchmittBitB6.Text == "N") value += 64;
            if (bProgramSchmittBitB5.Text == "N") value += 32;
            if (bProgramSchmittBitB4.Text == "N") value += 16;
            if (bProgramSchmittBitB3.Text == "N") value += 8;
            if (bProgramSchmittBitB2.Text == "N") value += 4;
            if (bProgramSchmittBitB1.Text == "N") value += 2;
            if (bProgramSchmittBitB0.Text == "N") value += 1;

            tbProgramSchmittHexB.Text = value.ToString("X2");
            tbProgramSchmittDecimalB.Text = value.ToString();
        }

        private void bProgramSchmittBitC_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "N") btn.Text = "S"; else btn.Text = "N";

            // update bit data
            byte value = 0;
            if (bProgramSchmittBitC7.Text == "N") value += 128;
            if (bProgramSchmittBitC6.Text == "N") value += 64;
            if (bProgramSchmittBitC5.Text == "N") value += 32;
            if (bProgramSchmittBitC4.Text == "N") value += 16;
            if (bProgramSchmittBitC3.Text == "N") value += 8;
            if (bProgramSchmittBitC2.Text == "N") value += 4;
            if (bProgramSchmittBitC1.Text == "N") value += 2;
            if (bProgramSchmittBitC0.Text == "N") value += 1;

            tbProgramSchmittHexC.Text = value.ToString("X2");
            tbProgramSchmittDecimalC.Text = value.ToString();
        }

        private void tbProgramSchmittHexA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramSchmittHexA.Text, updateProgramSchmittA, null);
        }

        private void tbProgramSchmittDecimalA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramSchmittDecimalA.Text, updateProgramSchmittA, null);
        }

        private void tbProgramSchmittHexB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramSchmittHexB.Text, updateProgramSchmittB, null);
        }

        private void tbProgramSchmittDecimalB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramSchmittDecimalB.Text, updateProgramSchmittB, null);
        }

        private void tbProgramSchmittHexC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbProgramSchmittHexC.Text, updateProgramSchmittC, null);
        }

        private void tbProgramSchmittDecimalC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbProgramSchmittDecimalC.Text, updateProgramSchmittC, null);
        }
        #endregion

        /*************************************************************
         * AUTOSCAN settings
         *************************************************************/
        #region AUTOSCAN BIT TEST settings
        private void bAutoscanBitA_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "1") btn.Text = "0"; else btn.Text = "1";

            // update bit data
            byte value = 0;
            if (bAutoscanMaskBitA7.Text == "1") value += 128;
            if (bAutoscanMaskBitA6.Text == "1") value += 64;
            if (bAutoscanMaskBitA5.Text == "1") value += 32;
            if (bAutoscanMaskBitA4.Text == "1") value += 16;
            if (bAutoscanMaskBitA3.Text == "1") value += 8;
            if (bAutoscanMaskBitA2.Text == "1") value += 4;
            if (bAutoscanMaskBitA1.Text == "1") value += 2;
            if (bAutoscanMaskBitA0.Text == "1") value += 1;

            updateAutoscanMaskA(value);
        }

        private void bAutoscanBitB_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "1") btn.Text = "0"; else btn.Text = "1";

            // update bit data
            byte value = 0;
            if (bAutoscanMaskBitB7.Text == "1") value += 128;
            if (bAutoscanMaskBitB6.Text == "1") value += 64;
            if (bAutoscanMaskBitB5.Text == "1") value += 32;
            if (bAutoscanMaskBitB4.Text == "1") value += 16;
            if (bAutoscanMaskBitB3.Text == "1") value += 8;
            if (bAutoscanMaskBitB2.Text == "1") value += 4;
            if (bAutoscanMaskBitB1.Text == "1") value += 2;
            if (bAutoscanMaskBitB0.Text == "1") value += 1;

            updateAutoscanMaskB(value);
        }

        private void bAutoscanBitC_Click(object sender, EventArgs e)
        {
            // get reference to textbox
            Button btn = (Button)sender;

            // toggle text
            if (btn.Text == "1") btn.Text = "0"; else btn.Text = "1";

            // update bit data
            byte value = 0;
            if (bAutoscanMaskBitC7.Text == "1") value += 128;
            if (bAutoscanMaskBitC6.Text == "1") value += 64;
            if (bAutoscanMaskBitC5.Text == "1") value += 32;
            if (bAutoscanMaskBitC4.Text == "1") value += 16;
            if (bAutoscanMaskBitC3.Text == "1") value += 8;
            if (bAutoscanMaskBitC2.Text == "1") value += 4;
            if (bAutoscanMaskBitC1.Text == "1") value += 2;
            if (bAutoscanMaskBitC0.Text == "1") value += 1;

            updateAutoscanMaskC(value);
        }

        private void tbAutoscanMaskHexA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbAutoscanMaskHexA.Text, updateAutoscanMaskA, null);
        }

        private void tbAutoscanMaskDecimalA_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbAutoscanMaskDecimalA.Text, updateAutoscanMaskA, null);
        }

        private void tbAutoscanMaskHexB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbAutoscanMaskHexB.Text, updateAutoscanMaskB, null);
        }

        private void tbAutoscanMaskDecimalB_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbAutoscanMaskDecimalB.Text, updateAutoscanMaskB, null);
        }

        private void tbAutoscanMaskHexC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue("0x" + tbAutoscanMaskHexC.Text, updateAutoscanMaskC, null);
        }

        private void tbAutoscanMaskDecimalC_Validating(object sender, CancelEventArgs e)
        {
            e.Cancel = !validateValue(tbAutoscanMaskDecimalC.Text, updateAutoscanMaskC, null);
        }
        #endregion

        #region AUTOSCAN PERIOD & FILTER

        private void tbAutoscanFilterCount_Validating(object sender, CancelEventArgs e)
        {
            int value;
            TextBox tbox = (TextBox)sender;

            if (!int.TryParse(tbox.Text, out value) || value < 0 || value > 255)
            {
                e.Cancel = true;
                MessageBox.Show("Value must be between 0 and 255", "Input Error", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                _EEVal[35] = (byte)value;
                EEShow();
            }
        }

        private void tbAutoscanScanPeriod_Validating(object sender, CancelEventArgs e)
        {
            int value;
            TextBox tbox = (TextBox)sender;

            if (!int.TryParse(tbox.Text, out value) || value < 0 || value > 65535)
            {
                e.Cancel = true;
                MessageBox.Show("Value must be between 0 and 65535", "Input Error", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                _EEVal[36] = (byte)(value % 256);
                _EEVal[37] = (byte)(value / 256);
                EEShow();
            }
        }
        #endregion

        /*************************************************************
         * DATA ENTRY VALIDATION & UPDATE
         *************************************************************/
        private void tbAutoscanIP_Validating(object sender, CancelEventArgs e)
        {
            byte value;

            TextBox tbox = (TextBox)sender;

            if (!byte.TryParse(tbox.Text, out value) || value < 0 || value > 255)
            {
                e.Cancel = true;
                MessageBox.Show("Value must be between 0 and 255", "Input Error", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                if (byte.TryParse(tbAutoscanIP1.Text, out value))
                    _EEVal[44] = value;
                if (byte.TryParse(tbAutoscanIP2.Text, out value))
                    _EEVal[45] = value;
                if (byte.TryParse(tbAutoscanIP3.Text, out value))
                    _EEVal[46] = value;
                if (byte.TryParse(tbAutoscanIP4.Text, out value))
                    _EEVal[47] = value;

                EEShow();
            }
        }

        private void tbAutoscanTargetPort_Validating(object sender, CancelEventArgs e)
        {
            int value;
            TextBox tbox = (TextBox)sender;

            if (!int.TryParse(tbox.Text, out value) || value < 0 || value > 65535)
            {
                e.Cancel = true;
                MessageBox.Show("Value must be between 0 and 65535", "Input Error", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                _EEVal[48] = (byte)(value % 256);
                _EEVal[49] = (byte)(value / 256);
                EEShow();
            }
        }

        private void tbAutoscanMAC_Validating(object sender, CancelEventArgs e)
        {
            byte value;
            TextBox tbox = (TextBox)sender;

            if (!byte.TryParse(tbox.Text, System.Globalization.NumberStyles.HexNumber, null, out value) || value < 0 || value > 255)
            {
                e.Cancel = true;
                MessageBox.Show("Value must be between 00 and FF", "Input Error", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                if (byte.TryParse(tbAutoscanMAC1.Text, System.Globalization.NumberStyles.HexNumber, null, out value))
                    _EEVal[38] = value;
                if (byte.TryParse(tbAutoscanMAC2.Text, System.Globalization.NumberStyles.HexNumber, null, out value))
                    _EEVal[39] = value;
                if (byte.TryParse(tbAutoscanMAC3.Text, System.Globalization.NumberStyles.HexNumber, null, out value))
                    _EEVal[40] = value;
                if (byte.TryParse(tbAutoscanMAC4.Text, System.Globalization.NumberStyles.HexNumber, null, out value))
                    _EEVal[41] = value;
                if (byte.TryParse(tbAutoscanMAC5.Text, System.Globalization.NumberStyles.HexNumber, null, out value))
                    _EEVal[42] = value;
                if (byte.TryParse(tbAutoscanMAC6.Text, System.Globalization.NumberStyles.HexNumber, null, out value))
                    _EEVal[43] = value;

                EEShow();
            }
        }

        private void bAutoscanTargetMe_Click(object sender, EventArgs e)
        {
            _AutoLoadTarget = true;
            if (_deviceEP != null)
            {
                byte[] dgram = { (byte)'%' };
                _udp.Send(dgram, dgram.Length, _deviceEP);
            }
        }

        private void cbAutoscanMaskEnable_CheckedChanged(object sender, EventArgs e)
        {
            if (!_ReadingEE)
            {

                if (cbAutoscanMaskEnable.Checked)
                    _EEVal[10] = (byte)(_EEVal[10] ^ 4);
                else
                    _EEVal[10] = (byte)(_EEVal[10] | 4);
            }

            EEShow();
        }

        private void tbFixedIP_Validating(object sender, CancelEventArgs e)
        {
            byte value;

            TextBox tbox = (TextBox)sender;

            if (!byte.TryParse(tbox.Text, out value) || value < 0 || value > 255)
            {
                e.Cancel = true;
                MessageBox.Show("Value must be between 0 and 255", "Input Error", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                if (byte.TryParse(tbFixedIP1.Text, out value))
                    _EEVal[12] = value;
                if (byte.TryParse(tbFixedIP2.Text, out value))
                    _EEVal[13] = value;
                if (byte.TryParse(tbFixedIP3.Text, out value))
                    _EEVal[14] = value;
                if (byte.TryParse(tbFixedIP4.Text, out value))
                    _EEVal[15] = value;

                EEShow();
            }
        }

        private void tbMaskIP_Validating(object sender, CancelEventArgs e)
        {
            byte value;

            TextBox tbox = (TextBox)sender;

            if (!byte.TryParse(tbox.Text, out value) || value < 0 || value > 255)
            {
                e.Cancel = true;
                MessageBox.Show("Value must be between 0 and 255", "Input Error", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                if (byte.TryParse(tbMaskIP1.Text, out value))
                    _EEVal[50] = value;
                if (byte.TryParse(tbMaskIP2.Text, out value))
                    _EEVal[51] = value;
                if (byte.TryParse(tbMaskIP3.Text, out value))
                    _EEVal[52] = value;
                if (byte.TryParse(tbMaskIP4.Text, out value))
                    _EEVal[53] = value;

                EEShow();
            }
        }

        private void tbGwayIP_Validating(object sender, CancelEventArgs e)
        {
            byte value;

            TextBox tbox = (TextBox)sender;

            if (!byte.TryParse(tbox.Text, out value) || value < 0 || value > 255)
            {
                e.Cancel = true;
                MessageBox.Show("Value must be between 0 and 255", "Input Error", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                if (byte.TryParse(tbGwayIP1.Text, out value))
                    _EEVal[54] = value;
                if (byte.TryParse(tbGwayIP2.Text, out value))
                    _EEVal[55] = value;
                if (byte.TryParse(tbGwayIP3.Text, out value))
                    _EEVal[56] = value;
                if (byte.TryParse(tbGwayIP4.Text, out value))
                    _EEVal[57] = value;

                EEShow();
            }
        }

        private void tbProgrammablePort_Validating(object sender, CancelEventArgs e)
        {
            int value;
            TextBox tbox = (TextBox)sender;

            if (!int.TryParse(tbox.Text, out value) || value < 0 || value > 65535)
            {
                e.Cancel = true;
                MessageBox.Show("Value must be between 0 and 65535", "Input Error", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                _EEVal[58] = (byte)(value % 256);
                _EEVal[59] = (byte)(value / 256);
                EEShow();
            }
        }

        private void cbEnableFixedIP_CheckedChanged(object sender, EventArgs e)
        {
            if (!_ReadingEE)
            {
                if (cbEnableFixedIP.Checked)
                    _EEVal[10] = (byte)(_EEVal[10] ^ 1);
                else
                    _EEVal[10] = (byte)(_EEVal[10] | 1);
            }
            EEShow();
        }

        private void MainForm_Activated(object sender, EventArgs e)
        {
            if (!_scanned)
            {

                _scanned = true;
                Module_Scan(null, null);
            }
        }

        private void tmrCheckAvailable_Tick(object sender, EventArgs e)
        {
            if (_udp == null)
                return;
            else
                waitForData(10);

        }

        private void miDebug_Click(object sender, EventArgs e)
        {
            debug.Visible = !debug.Visible;
        }

        private void TextBox_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Decimal)
            {
                TextBox tbx = (TextBox)sender;
                tbx.Parent.SelectNextControl(tbx, true, true, true, true);
                e.Handled = true;
            }
        }

        private void pbWriteStatus_Click(object sender, EventArgs e)
        {

        }

        private void tbProgramDirectionDecimalA_TextChanged(object sender, EventArgs e)
        {

        }

        private void panel1_Paint(object sender, PaintEventArgs e)
        {
            
        }

        private void tsddModules_Click(object sender, EventArgs e)
        {

        }

        private void menuStrip1_ItemClicked(object sender, ToolStripItemClickedEventArgs e)
        {
            
        }
    }
}
